package org.jboss.seam.selectitems.jsf;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.faces.application.Application;
import javax.faces.component.UIComponent;
import javax.faces.component.UISelectItems;
import javax.faces.component.ValueHolder;
import javax.faces.context.FacesContext;
import javax.faces.el.ValueBinding;
import javax.faces.model.SelectItem;
import javax.persistence.Entity;

import org.jboss.cache.CacheException;
import org.jboss.seam.core.PojoCache;
import org.jboss.seam.log.Log;
import org.jboss.seam.log.LogImpl;
import org.jboss.seam.selectitems.SelectItemsConfig;

import com.sun.facelets.FaceletContext;
import com.sun.facelets.tag.MetaRuleset;
import com.sun.facelets.tag.TagAttribute;
import com.sun.facelets.tag.TagException;
import com.sun.facelets.tag.jsf.ComponentConfig;
import com.sun.facelets.tag.jsf.ComponentHandler;

public class SelectItemsComponentHandler extends ComponentHandler {

	private Log log = new LogImpl(SelectItemsComponentHandler.class);
	private TagAttribute cacheKeyAttribute;
	private TagAttribute cacheRegionAttribute;
	private TagAttribute cacheAttribute;
	private TagAttribute varAttribute;
	private TagAttribute noSelectionLabelAttribute;
	private TagAttribute hideNoSelectionAttribute;
	private TagAttribute keyAttribute;
	
	private String cacheKey;

	public SelectItemsComponentHandler(ComponentConfig config) {
		super(config);
		this.cacheKeyAttribute = this.getAttribute("cacheKey");
		this.cacheRegionAttribute = this.getAttribute("cacheRegion");
		this.cacheAttribute = this.getAttribute("cache");
		this.varAttribute = this.getAttribute("var");
		this.noSelectionLabelAttribute = this.getAttribute("noSelectionLabel");
		this.hideNoSelectionAttribute = this.getAttribute("hideNoSelection");
		this.keyAttribute = this.getAttribute("key");
		}
	
	private boolean addConverters(UIComponent parent, FaceletContext ctx) {
		// Only add a converter if the user hasn't asked key as the 'value'
		if (keyAttribute == null) {
			if (parent instanceof ValueHolder) {
				ValueHolder valueHolder = (ValueHolder) parent;
				// Does an existing converter exist?
				ValueBinding vb = parent.getValueBinding("value");
				if (valueHolder.getConverter() != null) {
					// Ensure the converter implements the SelectItems converter
					if (!(valueHolder.getConverter() instanceof SelectItemsConverter)) {
						throw new TagException(this.tag, "Converter must implement SelectItemsConverter: " + valueHolder.getConverter() );
					}
				} else if (vb != null && vb.getType(ctx.getFacesContext()) != null) {
					Application application = ctx.getFacesContext().getApplication();
					Class<?> clazz = vb.getType(ctx.getFacesContext());
					if (clazz.isEnum()) {
						valueHolder.setConverter(application
								.createConverter("org.jboss.seam.EnumConverter"));
						return true;
					} else if (clazz.isAnnotationPresent(Entity.class)) {
						valueHolder.setConverter(application
								.createConverter("org.jboss.seam.EntityConverter"));
						return true;
					}
				} else {
					log.debug("No converters attached");
				}
			} else {
				throw new TagException(this.tag,
						"parent not instance of ValueHolder: " + parent);
			}
			return false;
		} else {
			return false;
		}
	}

	private boolean addNoSelectionLabel(UISelectItems cmp, FaceletContext ctx, List<SelectItem> selectItems, UIComponent parent) {
		boolean add = true;
		if (getHideNoSelection(ctx) && getValue(parent) != null){
			add = false;
		}
		if (add && getNoSelectionLabel(ctx) != null) {
			SelectItem s = new SelectItem(SelectItemsConfig.NO_SELECTION_VALUE, getMessage(getNoSelectionLabel(ctx), ctx));
			selectItems.add(s);
			return true;
		} else {
			return false;
		}
	}

	private void cacheValue(UISelectItems cmp, FaceletContext ctx) {
		try {
			PojoCache.instance().put(getCacheRegion(ctx), getCacheKey(cmp, ctx), cmp.getValue());
		} catch (CacheException ce) {
			log.error("error accessing cache", ce);
		}
	}
	
	@Override
	protected MetaRuleset createMetaRuleset(Class type) {
		return super.createMetaRuleset(type).ignore("cacheKey").ignore(
				"cacheRegion").ignore("cache").ignore("var").ignore("noSelectionLabel").ignore("useMessageBundle").ignore("hideNoSelection");
	}
	
	private void createSelectItems(UISelectItems cmp, UIComponent parent, FaceletContext ctx) {
		if (cmp.getValue() instanceof SelectItemsListModel) {
			// Fine, the value holds as is
		} else {
			// This is @Out style
			Object value = cmp.getValue();
			if (value instanceof Iterable) {
				List<SelectItem> selectItems = new ArrayList<SelectItem>();
				addNoSelectionLabel(cmp, ctx, selectItems, parent);
				for (Object o : (Iterable) value) {
					initVar(ctx.getFacesContext(), o);
					String label = resolveAndInterpolateProperty("label", ctx, cmp);
					Boolean disabled = resolvePropertyAsBoolean("disabled", ctx, cmp);
					String key = resolveProperty("key", ctx, cmp);
					selectItems.add(new SelectItem(keyAttribute == null ? o : key, label, "", disabled));
					destroyVar(ctx.getFacesContext());
				}
				cmp.setValue(selectItems);
			} else {
				throw new IllegalArgumentException("value must be iteratable");
			}
		}

	}
	private void destroyVar(FacesContext facesContext) {
		facesContext.getExternalContext().getRequestMap().remove(
				varAttribute.getValue());
	}
	
	private Boolean getCache(FaceletContext ctx) {
		if (cacheAttribute != null) {
			return cacheAttribute.getBoolean(ctx);
		} else{
			return SelectItemsConfig.instance().isCache();
		}
	}
	
	private String getCacheKey(UISelectItems cmp,
			FaceletContext ctx) {
		if (cacheKey == null) {
			if (cacheKeyAttribute != null
					&& !cacheKeyAttribute.getValue(ctx).equals("")) {
				this.cacheKey=  "org.jboss.seam.selectitems.cache."
						+ cacheKeyAttribute.getValue(ctx);
			} else if (getCache(ctx)) {
				// Work out cache key from value attribute
				ValueBinding vb = cmp.getValueBinding("value");
				String value = vb.getExpressionString();
				if (value.startsWith("#{") && value.endsWith("}")) {
					this.cacheKey = value.substring(2, value.length() - 1);
				}
			}
		}
		return cacheKey;
	}
	
	private String getCacheRegion(FaceletContext ctx) {
		if (cacheRegionAttribute != null
				&& !cacheRegionAttribute.getValue(ctx).equals("")) {
			return cacheRegionAttribute.getValue(ctx);
		} else {
			return SelectItemsConfig.instance().getCacheRegion();
		}
	}

	@SuppressWarnings("unchecked")
	private List<SelectItem> getFromCache(UISelectItems cmp, FaceletContext ctx) {
		try {
			return (List<SelectItem>) PojoCache.instance().get(getCacheRegion(ctx), getCacheKey(cmp, ctx));
		} catch (CacheException ce) {
			log.error("error accessing cache", ce);
			return null;
		}
	}

	private Boolean getHideNoSelection(FaceletContext ctx) {
		if (hideNoSelectionAttribute != null) {
			return hideNoSelectionAttribute.getBoolean(ctx);
		} else {
			return SelectItemsConfig.instance().isHideNoSelection();
		}
	}

	private String getMessage(String message, FaceletContext ctx) {
		return message;
	}

	private String getNoSelectionLabel(FaceletContext ctx) {
		if (noSelectionLabelAttribute != null) {
			return noSelectionLabelAttribute.getValue();
		} else {
			return SelectItemsConfig.instance().getNoSelectionLabel();
		}
	}

	private Object getValue(UIComponent cmp) {
		if (cmp instanceof ValueHolder) {
			ValueHolder valueHolder = (ValueHolder) cmp;
			return valueHolder.getValue();
		} else {
			throw new TagException(this.tag, "parent component must implement ValueHolder: " + cmp);
		}
	}

	@SuppressWarnings("unchecked")
	private void initVar(FacesContext facesContext, Object varValue) {
		if (varAttribute == null) {
			throw new TagException(this.tag, "var attribute must be set");
		}
		facesContext.getExternalContext().getRequestMap().put(
				varAttribute.getValue(), varValue);
	}
	
	private boolean isCacheEnabled(UISelectItems cmp, FaceletContext ctx) {
		return getCacheKey(cmp, ctx) != null;
	}
	
	@Override
	protected void onComponentCreated(FaceletContext ctx, UIComponent c, UIComponent parent) {
		super.onComponentCreated(ctx, c, parent);
		if (c instanceof UISelectItems) {
			addConverters(parent, ctx);
			UISelectItems cmp = (UISelectItems) c;
			if (isCacheEnabled(cmp, ctx)) {
				Object cachedContent = getFromCache(cmp, ctx);
				if (cachedContent != null) {
					cmp.setValue(cachedContent);
				} else {
					// Execute value binding now and place content in cache
					createSelectItems(cmp, parent, ctx);
					cacheValue(cmp, ctx);
				}
			} else {
				createSelectItems(cmp, parent, ctx);
			}
		}
	}
	
	private String resolveAndInterpolateProperty(String property,
			FaceletContext ctx, UIComponent cmp) {
		String resolvedProperty = resolveProperty(property, ctx, cmp);
		return getMessage(resolvedProperty, ctx);
	}
	
	private String resolveProperty(String property,
			FaceletContext ctx, UIComponent cmp) {
		Object propertyObject = cmp.getAttributes().get(property);
		if (propertyObject != null) {
			return propertyObject.toString();
		}
		ValueBinding vb = cmp.getValueBinding(property);
		if (vb != null) {
			Object vbObject = vb.getValue(ctx.getFacesContext());
			return vbObject == null ? null : vbObject.toString();
			
		}
		return null;
	}
	
	private Boolean resolvePropertyAsBoolean(String property,
			FaceletContext ctx, UIComponent cmp) {
		String propertyValue = resolveProperty(property, ctx, cmp);
		if (propertyValue != null) {
			return propertyValue.equals("true");
		} else {
			return false;
		}
	}

}
